function melody_bayes(wavfile,transfile,step)

% MELODY_BAYES Melody estimation via a loudness-weighted Bayesian harmonic
% model.
%
% melody_bayes(wavfile,transfile)
% melody_bayes(wavfile,transfile,step)
%
% Inputs:
% wavfile: input WAV filename
% transfile: output f0 filename
% step: stepsize in s (default: 0.0223)
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Copyright 2004-2005 Emmanuel Vincent (Queen Mary University of London)
% This software is distributed under the terms of the GNU Public License
% version 3 (http://www.gnu.org/licenses/gpl.txt)
% If you find it useful, please cite the following reference:
% Emmanuel Vincent, "Predominant-F0 estimation using Bayesian harmonic
% waveform models", in Proc. MIREX, 2005.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%lecture du fichier son
[x,fs]=wavread(wavfile);
x=resample(x,22050,fs);

%parametres fixes manuellement
w0=single(.5); w1=single(1); w2=single(1); wt=1;
nallowed=single(5); %nombre maximal de notes testees
factor_guess=1;  %weighting pour la creation des hypotheses de notes (89% de perf maximale avec nallowed=5)
factor_final=.5;  %weighting pour la selection de la meilleure hypothese

N=single(1024);
if nargin<3,
    S=N/2;
else
    S=22050*step;
end
x=single([zeros(N/2,1);x;zeros(N/2,1)]);
T=single(length(x)); x=norm_sound(x(:));
win=hanning(N);
nfram=floor((T-N)/S)+1;

firstnote=single(45); lastnote=single(84); nbnotes=lastnote-firstnote+1;
maxharm=single(60);

%recuperation des parametres du modele
load('melody_model.mat');
amcen=single(model.amcen); amvar=single(model.amvar); amean=single(model.amean); anvar=single(model.anvar);
fmcen=single(model.fmcen); fmvar=single(model.fmvar);
tnotnot=single(model.tnotnot); tnote=single(model.tnote); rtnote=single(model.rtnote);
nvar=single(.08);

%calcul de la matrice de lissage Bark et de la reponse de l'oreille externe
barkmat=zeros(N,'single');
f=[0:1/N:.5,.5-1/N:-1/N:1/N];
e=9.26*log(96.4*f+1);
a=213*exp(-e/9.26);
for b=1:N,
    ff=abs(sinc(2*pi*(f-f(b))*a(b))+.5*sinc(2*pi*(f-f(b))*a(b)+1)+.5*sinc(2*pi*(f-f(b))*a(b)-1));
    barkmat(b,:)=ff.*ff;
end
earfilt=single(10.^(earfilter3(f).'/10));
clear f e a b ff;

%tabulation de la fonction d'integration des phases
ntab=1201;
curv_tab=single(logspace(-9,3,ntab));
int_tab=zeros(ntab,1,'single');
nsampl=1000;
points=2*pi/nsampl*(-nsampl/2+1:nsampl/2);
for h=1:ntab,
    loglik=-2*curv_tab(h)*sin(points).^2;
    int_tab(h)=log(sum(exp(loglik))*2*pi/nsampl);
end
clear ntab nsampl points loglik;

%calcul de la probabilite des f0s les plus probables sur chaque trame
loglikmat=-inf(nbnotes,nfram,'single'); f0mat=nan(nbnotes,nfram,'single');
for t=1:nfram,
    y=x(round((t-1)*S)+1:round((t-1)*S)+N).*win;    
    %preprocessing et initialisation
    f0_guess=yin_notes(y,earfilt,barkmat,factor_guess);
    notes_guess=round(log(f0_guess*22050/440)/log(2)*12+70-firstnote);
    [notes_guess,keep]=list(notes_guess); f0_guess=f0_guess(keep);
    keep=single(find((notes_guess >= 1) & (notes_guess <= nbnotes)));
    nbkeep=min(nallowed,length(keep)); keep=keep(1:nbkeep);
    notes=notes_guess(keep);  %ces valeurs de pitch discret sont maintenant toutes differentes
    ampl=cell(nbkeep,1); phal=cell(nbkeep,1); am=zeros(nbkeep,1,'single'); f0=f0_guess(keep);
    loglik=-inf(nbkeep,1,'single'); up=ones(nbkeep,1,'single');
    for h=1:nbkeep,
        [ampl{h},phal{h},am(h)]=init(notes(h),f0(h),y,maxharm,amcen,amvar,amean,anvar);
    end
    %reestimation
    maxiter=single(50); thresh=single(.1); rthresh=single(.25); converged=single(0); iter=single(0); prevloglik=-inf(nbkeep,1,'single');
    while (iter <= maxiter) && ~converged,
        for h=single(find(up)).',
            [ampl{h},phal{h},am(h),f0(h),loglik(h)]=update(ampl{h},phal{h},am(h),f0(h),notes(h),y,earfilt,barkmat,factor_final,curv_tab,int_tab,nvar,w0,w1,w2,maxharm,amcen,amvar,amean,anvar,fmcen,fmvar);
        end
        logliktop=max([loglikmat(1,t);loglik]);
        up=isinf(prevloglik) | (abs(loglik-prevloglik) > max(thresh,rthresh*(logliktop-prevloglik)));
        up=up & ~isinf(loglik);
        converged=(iter > maxiter) | all(~up);
        prevloglik=loglik; iter=iter+1;
    end
    loglikmat(notes,t)=loglik;
    f0mat(notes,t)=f0;
    if all(isinf(loglikmat(notes,t))),
        loglikmat(:,t)=0;
        f0mat(:,t)=0;
    end
end

%lissage par chaine de Markov cachee
%initialisation
logiprob=wt*log(tnote);    %probas initiales
loglikv=loglikmat(:,1)+logiprob.';
psi=zeros(nbnotes,nfram-1);
%algorithme de Viterbi
tprob=tnotnot*eye(nbnotes);
for n=1:nbnotes,
    rtprob=[fliplr(rtnote(1:n-1)),rtnote(1:nbnotes-n+1)]; rtprob=rtprob/sum(rtprob);
    tprob(:,n)=tprob(:,n)+(1-tnotnot)*rtprob.';
end
logtprob=wt*log(tprob);    
for t=2:nfram,
    prevloglikv=loglikv;
    for n=1:nbnotes,
        [loglikv(n),psi(n,t-1)]=max(prevloglikv+logtprob(n,:).');
    end
    loglikv=loglikv+loglikmat(:,t);
end
%backward tracking
f0=zeros(1,nfram); path=zeros(1,nfram);
[val,path(nfram)]=max(loglikv);
f0(nfram)=f0mat(path(nfram),nfram)*22050;
for t=nfram-1:-1:1,
    path(t)=psi(path(t+1),t);
    f0(t)=f0mat(path(t),t)*22050;
end

%ecriture du fichier transcription
time=(0:S:(nfram-1)*S)/22050;
txt='';
for t=1:nfram,
    txt=[txt num2str(time(t)) char(9) num2str(f0(t)) char(10)];
end
fid=fopen(transfile,'w');
fprintf(fid,'%c',txt);
fclose(fid);

return;



%routine d'initialisation des parametres locaux d'un etat
function [ampl,phal,am]=init(note,f0,y,maxharm,amcen,amvar,amean,anvar)

N=single(length(y)); win=hanning(N); winen=(N+1)*3/8;
firstnote=single(45); lastnote=single(84); pitch=firstnote:lastnote;

nharm=min(maxharm,floor(11025/440/2^((pitch(note)-69)/12)));
modelpha=2*pi*f0*(-N/2+.5:N/2-.5).'*(1:nharm); exppha=win*ones(1,nharm).*exp(-i*modelpha);
amplc=y.'*exppha/winen; ampl=2*abs(amplc).'; phal=angle(amplc).';
am=(sum(log(ampl)-amean(1:nharm,note))/anvar(note)+amcen(note)/amvar(note))/(nharm/anvar(note)+1/amvar(note));
%ampl=exp(am+amean(1:nharm,note));

return;



%routine de calcul des parametres locaux d'un etat
function [ampl,phal,am,f0,loglik]=update(ampl,phal,am,f0,note,y,earfilt,barkmat,factor,curv_tab,int_tab,nvar,w0,w1,w2,maxharm,amcen,amvar,amean,anvar,fmcen,fmvar)

warning off;

N=single(length(y));
win=hanning(N); winen=(N+1)*3/8;

firstnote=single(45); lastnote=single(84); pitch=firstnote:lastnote; nbnotes=lastnote-firstnote+1;
excit=barkmat*(abs(fft(y)).^2);
if factor==0,
    excit=max(excit,1e-9*max(excit));
    wei=1./excit;
elseif factor==.5,
    wei=earfilt.^.25./excit.^.75;
elseif factor==1,
    wei=earfilt.^.5./excit.^.5;
elseif factor==2,
    wei=earfilt;
end
nharm=min(maxharm,floor(11025/440/2^((pitch(note)-69)/12)));
modelpha=2*pi*f0*(-N/2+.5:N/2-.5).'*(1:nharm); exppha=win*ones(1,nharm).*exp(-i*modelpha);
expspec=fft(exppha); weien=sum(wei*ones(1,nharm).*real(conj(expspec).*expspec)).'/2;

%reestimation
%calcul du modele et de la vraisemblance
modelpha=2*pi*f0*(-N/2+.5:N/2-.5).'*(1:nharm)+ones(N,1)*phal.';
cospha=cos(modelpha); sinpha=sin(modelpha);
model=cospha*ampl.*win;
error=fft(y-model);
amperror=log(ampl)-am-amean(1:nharm,note);
loglik0=-.5*real(error'*(error.*wei))/nvar-.5*N*log(2*pi*nvar);
loglik1=-.5*amperror.'*amperror/anvar(note)-.5*nharm*log(2*pi*anvar(note))-nharm*log(2*pi);
loglik2=-.5*(am-amcen(note)).^2/amvar(note)-.5*log(2*pi*amvar(note))-.5*(log(f0)-fmcen(note)).^2/fmvar(note)-.5*log(2*pi*fmvar(note));
loglik=w0*loglik0+w1*loglik1+w2*loglik2;
%update de toutes les variables
lstep=single(.5);
%update de f0
dmodel=fft(-sinpha*(ampl.*(1:nharm).')*2*pi.*(-N/2+.5:N/2-.5).'.*win);
grad0=w0*real(error'*(dmodel.*wei))*f0/nvar; curv0=w0*real(dmodel'*(dmodel.*wei))*f0*f0/nvar;
grad2=-w2*(log(f0)-fmcen(note))/fmvar(note); curv2=w2/fmvar(note);
f0=f0*exp(lstep*(grad0+grad2)./(curv0+curv2));
%update des amplitudes
dmodel=fft(win*ones(1,nharm).*cospha);
grad0=w0*sum(real(conj(error.*wei)*ones(1,nharm).*dmodel)).'.*ampl/nvar; curv0=w0*(weien.*ampl.*ampl)/nvar; %curv0=w0*sum(real(wei*ones(1,nharm).*conj(dmodel).*dmodel)).'.*ampl.*ampl/nvar;
grad1=-w1*amperror/anvar(note); curv1=w1/anvar(note);
ampl=ampl.*exp(lstep*(grad0+grad1)./(curv0+curv1));
%update des phases
%exppha=win*ones(1,nharm).*(cospha-i*sinpha);
%amplc=y.'*exppha/winen; phal=angle(amplc).'+phal; phal=mod(phal+pi,2*pi)-pi;
dmodel=fft(-win*ones(1,nharm).*sinpha);
grad0=sum(real(conj(error.*wei)*ones(1,nharm).*dmodel)).'; curv0=weien.*ampl; %curv0=sum(real(wei*ones(1,nharm).*conj(dmodel).*dmodel)).'.*ampl;
phal=phal+lstep*grad0./curv0; phal=mod(phal+pi,2*pi)-pi;
%update de am
grad1=w1*sum(amperror)/anvar(note); curv1=w1*nharm/anvar(note);
grad2=-w2*(am-amcen(note))/amvar(note); curv2=w2/amvar(note);
am=am+lstep*(grad1+grad2)/(curv1+curv2);
%methode de Laplace
%recalcul de la valeur maximale
modelpha=2*pi*f0*(-N/2+.5:N/2-.5).'*(1:nharm)+ones(N,1)*phal.';
cospha=cos(modelpha);
model=cospha*ampl.*win;
error=fft(y-model);
amperror=log(ampl)-am-amean(1:nharm,note);
loglik0=-.5*real(error'*(error.*wei))/nvar-.5*N*log(2*pi*nvar);
loglik1=-.5*amperror.'*amperror/anvar(note)-.5*nharm*log(2*pi*anvar(note))-nharm*log(2*pi);
loglik2=-.5*(am-amcen(note)).^2/amvar(note)-.5*log(2*pi*amvar(note))-.5*(log(f0)-fmcen(note)).^2/fmvar(note)-.5*log(2*pi*fmvar(note));
loglikm=w0*loglik0+w1*loglik1+w2*loglik2;
%integration par rapport aux amplitudes et aux phases
dmodel=fft(win*ones(1,nharm).*cospha);
curv0=w0*(sum(real(wei*ones(1,nharm).*conj(dmodel).*dmodel)).'.*ampl-sum(real(conj(error.*wei)*ones(1,nharm).*dmodel)).').*ampl/nvar;
curv0=max(curv0,0);
curv1=w1/anvar(note);
curvampl=curv0+curv1;
intampl=.5*nharm*log(2*pi)-.5*sum(log(curvampl));
dmodel=fft(-win*ones(1,nharm).*sinpha); d2model=-fft(win*ones(1,nharm).*cospha);
curvphal=w0*(sum(real(wei*ones(1,nharm).*conj(dmodel).*dmodel)).'.*ampl-sum(real(conj(error.*wei)*ones(1,nharm).*d2model)).').*ampl/nvar;
curvphal=max(1e-9,curvphal);
intphal=sum(interp1(log(curv_tab),int_tab,log(curvphal)));
loglik=loglikm+intampl+intphal;

if isnan(loglik), loglik=-Inf; end;

warning backtrace;

return;



%routine de calcul des minima de la fonction de difference ponderee
function [freq,val]=yin_notes(y,earfilt,barkmat,factor)

%donne un vecteur vide sur les trames de silence complet

firstnote=single(45); lastnote=single(84);
ups=single(5);   %facteur d'upsampling de l'echelle MIDI
noterange=lastnote+1:-1/ups:firstnote-1; nbnotes=length(noterange);
dvect=single(22050/440./2.^((noterange-69)/12));

N=single(length(earfilt));
y=fft(y); y=y(1:N/2+1);
earfilt=earfilt(1:N/2+1);
barkmat=barkmat(1:N/2+1,1:N/2+1);

%calcul du poids frequentiel
excit=barkmat*(abs(y).^2);
if factor==0,
    excit=max(excit,1e-9*max(excit));
    wei=1./excit;
elseif factor==.5,
    wei=earfilt.^.25./excit.^.75;
elseif factor==1,
    wei=earfilt.^.5./excit.^.5;
elseif factor==2,
    wei=earfilt;
end

%calcul de la difference et de la difference normalisee
df=zeros(nbnotes,1,'single');
for d=1:nbnotes,
    errf=y.*(1-exp(2*1i*pi*(0:N/2).'/N*dvect(d)));
    df(d)=real((wei.*errf)'*errf);
end
df=df/real((y.*wei)'*y);

%calcul des minima
peaks=find((df(2:nbnotes-1) < df(1:nbnotes-2)) & (df(2:nbnotes-1) < df(3:nbnotes)))+1;
nbpeaks=length(peaks);
val=zeros(nbpeaks,1,'single'); freq=zeros(nbpeaks,1,'single');
for p=1:nbpeaks,
    coeffs=polyfit(dvect(peaks(p)-1:peaks(p)+1).',df(peaks(p)-1:peaks(p)+1),2);
    freq(p)=-2*coeffs(1)/coeffs(2);
    val(p)=polyval(coeffs,1/freq(p));
end

%classement par ordre croissant d'erreur residuelle
[val,ind]=sort(val);
freq=freq(ind);

return;


function [s,coeff]=norm_sound(s)

%normalise un son a une puissance maximale de -12 dB calculee sur des fenetres rectangulaires de taille 256

[T nchan]=size(s);
if (nchan > T), s=s.'; end;
en=reshape([sum(s.^2,2);zeros(ceil(T/256)*256-T,1)],256,ceil(T/256));
en=max(sum(en))/256;
coeff=.25/sqrt(en);
s=s*coeff;

return;


function filt=earfilter3(f)

%filt=earfilter3(f)
%
%renvoie l'attenuation de l'oreille moyenne sur un ensemble de frequences
%reduites quelconques
%
%FORMULE DE MOORE

freqs=[0,20,25,31.5,40,50,63,80,100,125,...
160,200,250,315,400,500,630,800,1000,1250,1600,...
2000,2500,3150,4000,5000,6300,8000,9000,10000,12500,15000,20000];
datamaf=[75.8,70.1,60.8,52.1,44.2,37.5,31.3,25.6,20.9,...
16.5,12.6,9.6,7.0,4.7,3.0,1.8,0.8,0.2,0.0,-0.5,...
-1.6,-3.2,-5.4,-7.8,-8.1,-5.3,2.4,11.1,12.8,12.2,7.4,17.8,17.8];

fm=freqs/22050;
filt=-interp1(fm,datamaf,f,'spline');

return;


function [l,ind]=list(a)

%l=a(1); ind=1;
%for t=2:length(a),
l=[]; ind=[];
for t=1:length(a),
if ~any(l == a(t)),
l=[l;a(t)]; ind=[ind;t];
end
end

return;
